import abc
import copy
import numbers
import collections.abc
from types import MappingProxyType

import numpy as np

__all__ = ["NDMeta", "NDMetaABC"]


class NDMetaABC(collections.abc.Mapping):
    """
    A sliceable object for storing metadata.

    Metadata can be linked to a data axis which causes it to be sliced when the
    standard Python numeric slicing API is applied to the object.
    Specific pieces of metadata can be obtain using the dict-like string slicing API.
    Metadata associated with an axis/axes must have the same length/shape as those axes.

    Parameters
    ----------
    meta: dict-like
        The names and values of metadata.

    key_comments: dict-like, optional
        Comments associated with any of the above pieces of metadata.

    axes: dict-like, optional
        The axis/axes associated with the metadata denoted by the keys.
        Metadata not included are considered not to be associated with any axis.
        Each axis value must be an iterable of `int`. An `int` itself is also
        acceptable if the metadata is associated with a single axis.
        The value of axis-assigned metadata in meta must be same length as
        number of associated axes (axis-aligned), or same shape as the associated
        data array's axes (grid-aligned).

    Notes
    -----
    **Axis-aware Metadata**
    There are two valid types of axis-aware metadata: axis-aligned and grid-aligned.
    Axis-aligned metadata gives one value per associated axis, while grid-aligned
    metadata gives a value for each data array element in the associated axes.
    Consequently, axis-aligned metadata has the same length as the number of
    associated axes, while grid-aligned metadata has the same shape as the associated
    axes. To avoid confusion, axis-aligned metadata that is only associated with one
    axis must be scalar or a string. Length-1 objects (excluding strings) are assumed
    to be grid-aligned and associated with a length-1 axis.

    **Slicing and Rebinning Axis-aware Metadata**
    Axis-aligned metadata is only considered valid if the associated axes are present.
    Therefore, axis-aligned metadata is only changed if an associated axis is dropped
    by an operation, e.g. slicing. In such a case, the value associated with the
    dropped axes is also dropped and hence lost.  If the axis of a 1-axis-aligned
    metadata value (scalar) is slicing away, the metadata key is entirely removed
    from the NDMeta object.

    Grid-aligned metadata is mirrors the data array, it is sliced following
    the same rules with one exception. If an axis is dropped by slicing, the metadata
    name is kept, but its value is set to the value at the row/column where the
    axis/axes was sliced away, and the metadata axis-awareness is removed. This is
    similar to how coordinate values are transferred to ``global_coords`` when their
    associated axes are sliced away.

    Note that because rebinning does not drop axes, axis-aligned metadata is unaltered
    by rebinning. By contrast, grid-aligned metadata must necessarily by affected by
    rebinning. However, how it is affected depends on the nature of the metadata and
    there is no generalized solution. Therefore, this class does not alter the shape
    or values of grid-aligned metadata during rebinning, but simply removes its
    axis-awareness.  If specific pieces of metadata have a known way to behave during
    rebinning, this can be handled by subclasses or mixins.
    """

    __ndcube_can_slice__: bool
    __ndcube_can_rebin__: bool

    @property
    @abc.abstractmethod
    def axes(self):
        """
        Mapping from metadata keys to axes with which they are associated.

        Metadata not associated with any axes need not be represented here.
        """

    @property
    @abc.abstractmethod
    def key_comments(self):
        """
        Mapping from metadata keys to associated comments.

        Metadata without a comment need not be represented here.
        """

    @property
    @abc.abstractmethod
    def data_shape(self):
        """
        The shape of the data with which the metadata is associated.
        """

    @abc.abstractmethod
    def add(self, name, value, key_comment=None, axes=None, overwrite=False):
        """
        Add a new piece of metadata to instance.

        Parameters
        ----------
        name: `str`
            The name/label of the metadata.

        value: Any
            The value of the metadata. If axes input is not None, this must have the
            same length/shape as those axes as defined by ``self.data_shape``.

        key_comment: `str` or `None`
            Any comment associated with this metadata. Set to None if no comment desired.

        axes: `int`, iterable of `int`, or `None`
            The axis/axes with which the metadata is linked. If not associated with any
            axis, set this to None.

        overwrite: `bool`, optional
            If True, overwrites the entry of the name name if already present.
        """

    @property
    @abc.abstractmethod
    def slice(self):
        """
        A helper class which, when sliced, returns a new NDMeta with axis- and grid-aligned metadata sliced.

        Example
        -------
        >>> sliced_meta = meta.slice[0:3, :, 2] # doctest: +SKIP
        """

    @abc.abstractmethod
    def rebin(self, rebinned_axes, new_shape):
        """
        Adjusts grid-aware metadata to stay consistent with rebinned data.
        """


class NDMeta(dict, NDMetaABC):
    # Docstring in ABC
    __ndcube_can_slice__ = True
    __ndcube_can_rebin__ = True

    def __init__(self, meta=None, key_comments=None, axes=None, data_shape=None):
        self._original_meta = meta
        if data_shape is None:
            self._data_shape = np.array([], dtype=int)
        else:
            self._data_shape = np.asarray(data_shape).astype(int)

        if meta is None:
            meta = {}
        super().__init__(meta.items())
        meta_keys = meta.keys()

        if key_comments is None:
            self._key_comments = {}
        else:
            if not set(key_comments.keys()).issubset(set(meta_keys)):
                raise ValueError(
                    "All comments must correspond to a value in meta under the same key."
                )
            self._key_comments = key_comments

        if axes is None:
            self._axes = {}
        else:
            axes = dict(axes)
            if not set(axes.keys()).issubset(set(meta_keys)):
                raise ValueError(
                    "All axes must correspond to a value in meta under the same key.")
            self._axes = {key:self._sanitize_axis_value(axis, meta[key], key)
                          for key, axis in axes.items()}

    def _sanitize_axis_value(self, axis, value, key):
        axis_err_msg = ("Values in axes must be an integer or iterable of integers giving "
                        f"the data axis/axes associated with the metadata. axis = {axis}.")
        if isinstance(axis, numbers.Integral):
            axis = (axis,)
        if len(axis) == 0:
            return ValueError(axis_err_msg)
        # Verify each entry in axes is an iterable of ints or a scalar.
        if not (isinstance(axis, collections.abc.Iterable)
                and all(isinstance(i, numbers.Integral) for i in axis)):
            return ValueError(axis_err_msg)
        # If metadata's axis/axes include axis beyond current data shape, extend it.
        data_shape = self.data_shape
        if max(axis) >= len(data_shape):
            data_shape = np.concatenate((data_shape,
                                         np.zeros(max(axis) + 1 - len(data_shape), dtype=int)))
        # Check whether metadata is compatible with data shape based on shapes
        # of metadata already present.
        axis = np.asarray(axis)
        if _not_scalar(value):
            axis_shape = data_shape[axis]
            if not _is_axis_aligned(value, axis_shape):
                # If metadata corresponds to previously unconstrained axis, update data_shape.
                idx0 = axis_shape == 0
                if idx0.any():
                    axis_shape[idx0] = np.array(_get_metadata_shape(value))[idx0]
                    data_shape[axis] = axis_shape
                # Confirm metadata is compatible with data shape.
                if not _is_grid_aligned(value, axis_shape):
                    raise ValueError(
                        f"{key} must have same shape {tuple(data_shape[axis])} "
                        f"as its associated axes {axis}, ",
                        f"or same length as number of associated axes ({len(axis)}). "
                        f"Has shape {value.shape if hasattr(value, 'shape') else len(value)}")
        elif len(axis) != 1:
            raise ValueError("Scalar and str metadata can only be assigned to one axis. "
                             f"key = {key}; value = {value}; axes = {axis}")
        self._data_shape = data_shape
        return axis

    @property
    def key_comments(self):
        return self._key_comments

    @property
    def axes(self):
        return self._axes

    @property
    def data_shape(self):
        return self._data_shape

    @data_shape.setter
    def data_shape(self, new_shape):
        """
        Set data shape to new shape.

        Must agree with shapes of any axes already associated with metadata

        Parameters
        ----------
        new_shape: array-like
            The new shape of the data. Elements must of of type `int`.
        """
        new_shape = np.round(new_shape).astype(int)
        if (new_shape < 0).any():
            raise ValueError("new_shape cannot include negative numbers.")
        # Confirm input shape agrees with shapes of pre-existing metadata.
        old_shape = self.data_shape
        if len(new_shape) != len(old_shape) and len(self._axes) > 0:
            n_meta_axes = max([ax.max() for ax in self._axes.values()]) + 1
            old_shape = np.zeros(n_meta_axes, dtype=int)
            for key, ax in self._axes.items():
                old_shape[ax] = np.asarray(self[key].shape)
        # Axes of length 0 are deemed to be of unknown length, and so do not have to match.
        idx, = np.where(old_shape > 0)
        if len(idx) > 0 and (old_shape[idx] != new_shape[idx]).any():
            raise ValueError("new_shape not compatible with pre-existing metadata. "
                             f"old shape = {old_shape}, new_shape = {new_shape}")
        self._data_shape = new_shape

    def add(self, name, value, key_comment=None, axes=None, overwrite=False):
        # Docstring in ABC.
        if name in self.keys() and overwrite is not True:
            raise KeyError(f"'{name}' already exists. "
                           "To update an existing metadata entry set overwrite=True.")
        if key_comment is not None:
            self._key_comments[name] = key_comment
        if axes is not None:
            axes = self._sanitize_axis_value(axes, value, name)
            self._axes[name] = axes
            # Adjust data shape if not already set.
            axis_shape = self._data_shape[np.asarray(axes)]
            if _is_grid_aligned(value, axis_shape) and (self._data_shape[self._axes[name]] == 0).any():
                value_shape = np.asarray(value.shape)
                data_shape = self._data_shape
                # If new value represents axes not yet represented in Meta object,
                # add zero-length axes in their place to be filled in.
                if len(value_shape) > len(data_shape):
                    data_shape = np.concatenate(
                        (data_shape, np.zeros(len(value_shape) - len(data_shape), dtype=int)))
                idx_value, = np.where(data_shape[axes] == 0)
                data_shape[axes[idx_value]] = value_shape[idx_value]
                self._data_shape = data_shape
        elif name in self._axes:
            del self._axes[name]
        # This must be done after updating self._axes otherwise it may error.
        self.__setitem__(name, value)

    def __delitem__(self, name):
        if name in self._key_comments:
            del self._key_comments[name]
        if name in self._axes:
            del self._axes[name]
        super().__delitem__(name)

    def __setitem__(self, key, val):
        axis = self.axes.get(key, None)
        if axis is not None:
            if _not_scalar(val):
                axis_shape = tuple(self.data_shape[axis])
                if not _is_grid_aligned(val, axis_shape) and not _is_axis_aligned(val, axis_shape):
                    raise TypeError(
                        f"{key} is already associated with axis/axes {axis}. val must therefore "
                        f"must either have same length as number associated axes ({len(axis)}), "
                        f"or the same shape as associated data axes {tuple(self.data_shape[axis])}. "
                        f"val shape = {val.shape if hasattr(val, 'shape') else (len(val),)}\n"
                        "We recommend using the 'add' method to set values.")
        super().__setitem__(key, val)

    @property
    def original_meta(self):
        return MappingProxyType(self._original_meta)

    @property
    def slice(self):
        # Docstring in ABC.
        return _NDMetaSlicer(self)

    def rebin(self, bin_shape):
        """
        Adjusts axis-aware metadata to stay consistent with a rebinned `~ndcube.NDCube`.

        This is done by simply removing the axis-awareness of metadata associated with
        rebinned axes. The metadata itself is not changed or removed. This operation
        does not remove axis-awareness from metadata only associated with non-rebinned
        axes, i.e. axes whose corresponding entries in ``bin_shape`` are 1.

        Parameters
        ----------
        bin_shape : array-like
            The number of pixels in a bin in each dimension.
        """
        # Sanitize input
        bin_shape = np.round(bin_shape).astype(int)
        data_shape = self.data_shape
        bin_shape = bin_shape[:len(data_shape)]  # Drop info on axes not defined by NDMeta.
        if (np.mod(data_shape, bin_shape) != 0).any():
            raise ValueError("bin_shape must be integer factors of their associated axes.")
        # Remove axis-awareness from grid-aligned metadata associated with rebinned axes.
        rebinned_axes = set(np.where(bin_shape != 1)[0])
        new_meta = copy.deepcopy(self)
        null_set = set()
        for name, axes in self.axes.items():
            if (_is_grid_aligned(self[name], data_shape[axes])
                and set(axes).intersection(rebinned_axes) != null_set):
                del new_meta._axes[name]
        # Update data shape.
        new_meta._data_shape = new_meta._data_shape // bin_shape
        return new_meta


class _NDMetaSlicer:
    """
    Helper class to slice an NDMeta instance using a slicing item.

    Parameters
    ----------
    meta: `NDMetaABC`
        The metadata object to slice.
    """
    def __init__(self, meta):
        self.meta = meta

    def __getitem__(self, item):
        data_shape = self.meta.data_shape
        if len(data_shape) == 0:
            raise TypeError("NDMeta object does not have a shape and so cannot be sliced.")

        new_meta = copy.deepcopy(self.meta)
        naxes = len(data_shape)
        if isinstance(item, (numbers.Integral, slice)):
            item = [item]
        if len(item) < naxes:
            item = np.array(list(item) + [slice(None)] * (naxes - len(item)), dtype=object)
        elif len(item) > naxes:
            # If item applies to more axes than have been defined in NDMeta,
            # ignore items applying to those additional axes.
            item = np.array(item[:naxes])
        else:
            item = np.asarray(item)
        # Replace non-int item elements corresponding to length-0 axes
        # with slice(None) so data shape is not altered.
        idx = [not isinstance(i, numbers.Integral) and s == 0 for i, s in zip(item, data_shape)]
        idx = np.arange(len(idx))[idx]
        item[idx] = np.array([slice(None)] * len(idx))

        # Edit data shape and calculate which axis will be dropped.
        dropped_axes = np.zeros(naxes, dtype=bool)
        new_shape = new_meta.data_shape
        for i, axis_item in enumerate(item):
            if isinstance(axis_item, numbers.Integral):
                dropped_axes[i] = True
            elif isinstance(axis_item, slice):
                start = axis_item.start
                if start is None:
                    start = 0
                if start < 0:
                    start = data_shape[i] - start
                stop = axis_item.stop
                if stop is None:
                    stop = data_shape[i]
                if stop < 0:
                    stop = data_shape[i] - stop
                new_shape[i] = stop - start
            else:
                raise TypeError("Unrecognized slice type. "
                                "Must be an int, slice and tuple of the same.")
        kept_axes = np.invert(dropped_axes)
        new_meta._data_shape = new_shape[kept_axes]

        # Slice all metadata associated with axes.
        for key, value in self.meta.items():
            if (axis := self.meta.axes.get(key, None)) is None:
                continue
            drop_key = False
            # Calculate new axis indices.
            new_axis = np.asarray(list(
                set(axis).intersection(set(np.arange(naxes)[kept_axes]))
                ))
            if len(new_axis) == 0:
                new_axis = None
            else:
                cumul_dropped_axes = np.cumsum(dropped_axes)[new_axis]
                new_axis -= cumul_dropped_axes

            # Calculate sliced metadata values.
            axis_shape = tuple(self.meta.data_shape[axis])
            if _is_scalar(value):
                new_value = value
                # If scalar metadata's axes have been dropped, mark metadata to be dropped.
                if new_axis is None:
                    drop_key = True
            else:
                value_is_axis_aligned = _is_axis_aligned(value, axis_shape)
                if value_is_axis_aligned:
                    new_item = kept_axes[axis]
                else:
                    new_item = tuple(item[axis])
                # Slice metadata value.
                try:
                    new_value = value[new_item]
                except Exception:  # noqa: BLE001
                    # If value cannot be sliced by fancy slicing, convert it
                    # it to an array, slice it, and then if necessary, convert
                    # it back to its original type.
                    new_value = (np.asanyarray(value)[new_item])
                    if hasattr(new_value, "__len__"):
                        new_value = type(value)(new_value)
                # If axis-aligned metadata sliced down to length 1, convert to scalar.
                if value_is_axis_aligned and len(new_value) == 1:
                    new_value = new_value[0]
            # Overwrite metadata value with newly sliced version.
            if drop_key:
                del new_meta[key]
            else:
                new_meta.add(key, new_value, self.meta.key_comments.get(key, None), new_axis,
                             overwrite=True)

        return new_meta


def _not_scalar(value):
    return (
        (
         hasattr(value, "shape")
         or hasattr(value, "__len__")
        )
        and not
        (
         isinstance(value, str)
        ))


def _is_scalar(value):
    return not _not_scalar(value)


def _get_metadata_shape(value):
    return value.shape if hasattr(value, "shape") else (len(value),)

def _is_grid_aligned(value, axis_shape):
    if _is_scalar(value):
        return False
    value_shape = _get_metadata_shape(value)
    if value_shape != tuple(axis_shape):
        return False
    return True


def _is_axis_aligned(value, axis_shape):
    len_value = len(value) if _not_scalar(value) else 1
    return not _is_grid_aligned(value, axis_shape) and len_value == len(axis_shape)




import numpy as np
import pytest

from astropy.wcs import WCS

from ndcube import utils

ht_with_celestial = {
    'CTYPE4': 'HPLN-TAN', 'CUNIT4': 'deg', 'CDELT4': 1, 'CRPIX4': 0, 'CRVAL4': 0, 'NAXIS4': 1,
    'CNAME4': 'redundant axis', 'CROTA4': 0,
    'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
    'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
    'NAXIS2': 3,
    'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}

hm = {'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
      'NAXIS1': 4,
      'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5,
      'NAXIS2': 3,
      'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm = WCS(header=hm)

hm_reindexed_102 = {
    'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 10,
    'NAXIS2': 4,
    'CTYPE1': 'HPLT-TAN', 'CUNIT1': 'deg', 'CDELT1': 0.5, 'CRPIX1': 2, 'CRVAL1': 0.5, 'NAXIS1': 3,
    'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm_reindexed_102 = WCS(header=hm_reindexed_102)


@pytest.fixture
def axis_correlation_matrix():
    return _axis_correlation_matrix()


def _axis_correlation_matrix():
    return np.array([[True, True, False, False],
                     [True, True, False, False],
                     [False, False, True, False],
                     [True, False, False, True]], dtype=bool)


@pytest.fixture
def test_wcs():
    return WCSTest()


class WCSTest:
    def __init__(self):
        self.world_axis_physical_types = [
            'custom:pos.helioprojective.lon', 'custom:pos.helioprojective.lat', 'em.wl', 'time']
        self.axis_correlation_matrix = _axis_correlation_matrix()


def test_convert_between_array_and_pixel_axes():
    test_input = np.array([1, 4, -2])
    naxes = 5
    expected = np.array([3, 0, 1])
    output = utils.wcs.convert_between_array_and_pixel_axes(test_input, naxes)
    assert all(output == expected)


def test_pixel_axis_to_world_axes(axis_correlation_matrix):
    output = utils.wcs.pixel_axis_to_world_axes(0, axis_correlation_matrix)
    expected = np.array([0, 1, 3])
    assert all(output == expected)


def test_world_axis_to_pixel_axes(axis_correlation_matrix):
    output = utils.wcs.world_axis_to_pixel_axes(1, axis_correlation_matrix)
    expected = np.array([0, 1])
    assert all(output == expected)


def test_pixel_axis_to_physical_types(test_wcs):
    output = utils.wcs.pixel_axis_to_physical_types(0, test_wcs)
    expected = np.array(['custom:pos.helioprojective.lon',
                         'custom:pos.helioprojective.lat', 'time'])
    assert all(output == expected)


def test_physical_type_to_pixel_axes(test_wcs):
    output = utils.wcs.physical_type_to_pixel_axes('lon', test_wcs)
    expected = np.array([0, 1])
    assert all(output == expected)


@pytest.mark.parametrize(('test_input', 'expected'), [('wl', 2), ('em.wl', 2)])
def test_physical_type_to_world_axis(test_input, expected):
    world_axis_physical_types = ['custom:pos.helioprojective.lon',
                                 'custom:pos.helioprojective.lat', 'em.wl', 'time']
    output = utils.wcs.physical_type_to_world_axis(test_input, world_axis_physical_types)
    assert output == expected


def test_get_dependent_pixel_axes(axis_correlation_matrix):
    output = utils.wcs.get_dependent_pixel_axes(0, axis_correlation_matrix)
    expected = np.array([0, 1, 3])
    assert all(output == expected)


def test_get_dependent_array_axes(axis_correlation_matrix):
    output = utils.wcs.get_dependent_array_axes(3, axis_correlation_matrix)
    expected = np.array([0, 2, 3])
    assert all(output == expected)


def test_get_dependent_world_axes(axis_correlation_matrix):
    output = utils.wcs.get_dependent_world_axes(3, axis_correlation_matrix)
    expected = np.array([0, 1, 3])
    assert all(output == expected)


def test_get_dependent_physical_types(test_wcs):
    output = utils.wcs.get_dependent_physical_types("time", test_wcs)
    expected = np.array(['custom:pos.helioprojective.lon',
                         'custom:pos.helioprojective.lat', 'time'])
    assert all(output == expected)


def test_array_indices_for_world_objects(wcs_4d_t_l_lt_ln):
    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_t_l_lt_ln, None)
    assert len(array_indices) == 3
    assert array_indices == ((3,), (2,), (0, 1))

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_t_l_lt_ln, ('time',))
    assert len(array_indices) == 1
    assert array_indices == ((3,),)

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_t_l_lt_ln, ('time', 'em.wl'))
    assert len(array_indices) == 2
    assert array_indices == ((3,), (2,))

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_t_l_lt_ln, ('lat',))
    assert len(array_indices) == 1
    assert array_indices == ((0, 1),)


def test_array_indices_for_world_objects_2(wcs_4d_lt_t_l_ln):
    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_lt_t_l_ln, None)
    assert len(array_indices) == 3
    assert array_indices == ((0, 3), (2,), (1,))

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_lt_t_l_ln, ('lat',))
    assert len(array_indices) == 1
    assert array_indices == ((0, 3),)

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_lt_t_l_ln, ('lat', 'time'))
    assert len(array_indices) == 2
    assert array_indices == ((0, 3), (2,))

    array_indices = utils.wcs.array_indices_for_world_objects(wcs_4d_lt_t_l_ln, ('lon', 'time'))
    assert len(array_indices) == 2
    assert array_indices == ((0, 3), (2,))


def test_compare_wcs_physical_types(wcs_4d_t_l_lt_ln, wcs_3d_l_lt_ln):
    assert utils.wcs.compare_wcs_physical_types(wcs_4d_t_l_lt_ln, wcs_4d_t_l_lt_ln) is True
    assert utils.wcs.compare_wcs_physical_types(wcs_4d_t_l_lt_ln, wcs_3d_l_lt_ln) is False


def test_identify_invariant_axes(wcs_3d_l_lt_ln):
    source_wcs = wcs_3d_l_lt_ln

    target_wcs_header = wcs_3d_l_lt_ln.low_level_wcs.to_header().copy()
    target_wcs_header['CDELT2'] = 10
    target_wcs_header['CDELT3'] = 20
    target_wcs = WCS(header=target_wcs_header)

    invariant_axes = utils.wcs.identify_invariant_axes(source_wcs, target_wcs, (4, 4, 4))
    assert invariant_axes == [True, False, False]

    invariant_axes = utils.wcs.identify_invariant_axes(source_wcs, target_wcs, (4, 4, 4),
                                                       atol=1e-20, rtol=1e-20)
    assert invariant_axes == [False, False, False]




import numpy as np
import pytest

from astropy.nddata import StdDevUncertainty

from ndcube.utils.cube import propagate_rebin_uncertainties


@pytest.fixture
def stacked_pixel_data():
    return np.array([[[0, 2, 4], [12, 14, 16]],
                     [[1, 3, 5], [13, 15, 17]],
                     [[6, 8, 10], [18, 20, 22]],
                     [[7, 9, 11], [19, 21, 23]]], dtype=float) + 1


def test_propagate_rebin_uncertainties_mean(stacked_pixel_data):
    # Build inputs
    data = stacked_pixel_data
    mask = data < 0
    mask[3, 0, 0] = True
    uncertainty = StdDevUncertainty(data * 0.1)

    # Build expected output
    expected = np.sqrt((uncertainty.array**2).sum(axis=0)) / data.shape[0]
    expected[0, 0] = np.sqrt((uncertainty.array[:3, 0, 0]**2).sum()) / 3
    expected = StdDevUncertainty(expected)

    # Run function
    output = propagate_rebin_uncertainties(uncertainty, data, mask,
                                           np.mean, operation_ignores_mask=False)

    assert type(output) is type(expected)
    assert np.allclose(output.array, expected.array)


def test_propagate_rebin_uncertainties_use_masked(stacked_pixel_data):
    # Build inputs
    data = stacked_pixel_data
    mask = data < 0
    mask[3, 0, 0] = True
    uncertainty = StdDevUncertainty(data * 0.1)

    # Build expected output
    expected = np.sqrt((uncertainty.array**2).sum(axis=0)) / data.shape[0]
    expected = StdDevUncertainty(expected)

    # Run function
    output = propagate_rebin_uncertainties(uncertainty, data, mask,
                                           np.mean, operation_ignores_mask=True)

    assert type(output) is type(expected)
    assert np.allclose(output.array, expected.array)


def test_propagate_rebin_uncertainties_prod(stacked_pixel_data):
    # Build inputs
    data = stacked_pixel_data
    mask = data < 0
    uncertainty = StdDevUncertainty(data * 0.1)

    # Build expected output
    binned_data = data.prod(axis=0)
    # TODO: Why do I have to divide by a factor 2 here?
    expected = np.sqrt(((uncertainty.array / data)**2).sum(axis=0)) * binned_data / 2
    expected = StdDevUncertainty(expected)

    # Run function
    output = propagate_rebin_uncertainties(uncertainty, data, mask,
                                           np.prod, operation_ignores_mask=False)

    assert type(output) is type(expected)
    assert np.allclose(output.array, expected.array)


def test_propagate_rebin_uncertainties_nan(stacked_pixel_data):
    # Build inputs
    data = stacked_pixel_data
    data[3, 0, 0] = np.nan
    mask = data < 0
    uncertainty = StdDevUncertainty(data * 0.1)

    # Build expected output
    expected = np.sqrt((uncertainty.array**2).sum(axis=0)) / data.shape[0]
    expected[0, 0] = np.sqrt((uncertainty.array[:3, 0, 0]**2).sum()) / 3
    expected = StdDevUncertainty(expected)

    # Run function
    output = propagate_rebin_uncertainties(uncertainty, data, mask,
                                           np.nanmean, operation_ignores_mask=False)

    assert type(output) is type(expected)
    assert np.allclose(output.array, expected.array)

    # Test another code path that does the same thing.
    mask = None
    output = propagate_rebin_uncertainties(uncertainty, data, mask,
                                           np.nanmean, operation_ignores_mask=False)
    assert type(output) is type(expected)
    assert np.allclose(output.array, expected.array)



import copy
import numbers
import textwrap
import collections.abc

import numpy as np

import ndcube.utils.collection as collection_utils
from ndcube.utils.exceptions import warn_deprecated

__all__ = ["NDCollection"]


class NDCollection(dict):
    """
    Class for holding and manipulating an unordered collection of NDCubes or NDCubeSequences.

    Parameters
    ----------
    key_data_pairs: sequence of `tuple` of (`str`, `~ndcube.NDCube` or `~ndcube.NDCubeSequence`)
        The names and data cubes/sequences to held in the collection.

    aligned_axes: `tuple` of `int`, `tuple` of `tuple` of `int`, 'all', or None, optional
        Axes of each cube/sequence that are aligned in numpy order.
        If elements are int, then the same axis numbers in all cubes/sequences are aligned.
        If elements are tuples of ints, then there must be one tuple for every cube/sequence.
        Each element of each tuple gives the axes of each cube/sequence that are aligned.
        If 'all', all axes are aligned in natural order, i.e. the 0th axes of all cubes
        are aligned, as are the 1st, and so on.
        Default is `None`.

    meta: `dict`, optional
        General metadata for the overall collection.

    Example
    -------
    Say the collection holds two NDCubes, each of 3 dimensions.

    >>> aligned_axes = (1, 2)  # doctest:  +SKIP

    means that axis 1 (0-based counting) of cube0 is aligned with axis 1 of cube1,
    and axis 2 of cube0 is aligned with axis 2 of cube1.
    However, if

    >>> aligned_axes = ((0, 1), (2, 1))  # doctest: +SKIP

    then the first tuple corresponds to cube0 and the second with cube1.
    This is interpreted as axis 0 of cube0 is aligned with axis 2 of cube1 while
    axis 1 of cube0 is aligned with axis 1 of cube1.
    """

    def __init__(self, key_data_pairs, aligned_axes=None, meta=None, **kwargs):
        for key, _ in key_data_pairs:
            if isinstance(key, numbers.Number):
                warn_deprecated(
                    "Passing numerical keys to NDCollection is deprecated as they"
                    " lead to ambiguity when slicing the collection."
                )
        # Enter data and metadata into object.
        super().__init__(key_data_pairs)
        self.meta = meta

        # Convert aligned axes to required format.
        sanitize_inputs = kwargs.pop("sanitize_inputs", True)
        if aligned_axes is not None:
            keys, data = zip(*key_data_pairs)
            # Sanitize aligned axes unless hidden kwarg indicates not to.
            if sanitize_inputs:
                aligned_axes = collection_utils._sanitize_aligned_axes(keys, data, aligned_axes)
            else:
                aligned_axes = dict(zip(keys, aligned_axes))
        if kwargs:
            raise TypeError(
                f"__init__() got an unexpected keyword argument: '{list(kwargs.keys())[0]}'"
            )
        # Attach aligned axes to object.
        self._aligned_axes = aligned_axes
        if self._aligned_axes is None:
            self.n_aligned_axes = 0
        else:
            self.n_aligned_axes = len(self.aligned_axes[keys[0]])

    @property
    def aligned_axes(self):
        """
        The axes of each array that are aligned in numpy order.
        """
        return self._aligned_axes

    @property
    def _first_key(self):
        return list(self.keys())[0]

    def __str__(self):
        return (textwrap.dedent(f"""\
            NDCollection
            ------------
            Cube keys: {tuple(self.keys())}
            Number of Cubes: {len(self)}
            Aligned dimensions: {self.aligned_dimensions}
            Aligned physical types: {self.aligned_axis_physical_types}"""))

    def __repr__(self):
        return f"{object.__repr__(self)}\n{self!s}"

    @property
    def aligned_dimensions(self):
        """
        The lengths of all aligned axes.

        If there are no aligned axes, returns None.
        """
        if self.aligned_axes is not None:
            return np.asanyarray(self[self._first_key].shape, dtype=object)[
                np.array(self.aligned_axes[self._first_key])
            ]
        return None

    @property
    def aligned_axis_physical_types(self):
        """
        The physical types common to all members that are associated with each aligned axis.

        One tuple is returned for each axis as there can be more than one physical type
        associated with an aligned axis. If there are no physical types associated
        with an aligned that is common to all collection members, an empty tuple is
        returned for that axis.
        """
        if self.aligned_axes is None:
            return None
        # Get array axis physical types for each aligned axis for all members of collection.
        collection_types = [np.array(cube.array_axis_physical_types,
                                     dtype=object)[np.array(self.aligned_axes[name])]
                            for name, cube in self.items()]
        # Return physical types common to all members of collection for each axis.
        return [tuple(set.intersection(*[set(cube_types[i]) for cube_types in collection_types]))
                for i in range(self.n_aligned_axes)]

    def __getitem__(self, item):
        # There are two ways to slice:
        # by key or sequence of keys, i.e. slice out given cubes in the collection, or
        # by typical python numeric slicing API,
        # i.e. slice the each component cube along the aligned axes.

        # If item is single string, slicing is simple.
        if isinstance(item, str):
            return super().__getitem__(item)

        # If item is not a single string...
        # If item is a sequence, ensure strings and numeric items are not mixed.
        item_is_strings = False
        if isinstance(item, collections.abc.Sequence):
            item_strings = [isinstance(item_, str) for item_ in item]
            item_is_strings = all(item_strings)
            # Ensure strings are not mixed with slices.
            if (not item_is_strings) and (not all(np.invert(item_strings))):
                raise TypeError("Cannot mix keys and non-keys when indexing instance.")

        # If sequence is all strings, extract the cubes corresponding to the string keys.
        if item_is_strings:
            new_data = [self[_item] for _item in item]
            new_keys = item
            new_aligned_axes = tuple([self.aligned_axes[item_] for item_ in item])
            new_meta = copy.deepcopy(self.meta)

        # Else, the item is assumed to be a typical slicing item.
        # Slice each cube in collection using information in this item.
        # However, this can only be done if there are aligned axes.
        else:
            if self.aligned_axes is None:
                raise IndexError("Cannot slice unless collection has aligned axes.")
            # Derive item to be applied to each cube in collection and
            # whether any aligned axes are dropped by the slicing.
            collection_items, new_aligned_axes = self._generate_collection_getitems(item)
            # Apply those slice items to each cube in collection.
            new_data = [self[key][tuple(cube_item)]
                        for key, cube_item in zip(self, collection_items)]
            # Since item is not strings, no cube in collection is dropped.
            # Therefore the collection keys remain unchanged.
            new_keys = list(self.keys())
            # Slice meta if sliceable
            if hasattr(self.meta, "__ndcube_can_slice__") and self.meta.__ndcube_can_slice__:
                # Convert negative indices to positive indices as they are not supported by NDMeta.slice
                sanitized_item = copy.deepcopy(item)
                aligned_shape = self.aligned_dimensions
                if isinstance(item, numbers.Integral):
                    if item < 0:
                        sanitized_item = int(self.aligned_dimensions[0] + item)
                elif isinstance(item, slice):
                    if (item.start is not None and item.start < 0) or (item.stop is not None and item.stop < 0):
                        new_start = aligned_shape[0] + item.start if item.start < 0 else item.start
                        new_stop = aligned_shape[0] + item.stop if item.stop < 0 else item.stop
                        sanitized_item = slice(new_start, new_stop)
                else:
                    sanitized_item = list(sanitized_item)
                    for i, ax_it in enumerate(item):
                        if isinstance(ax_it, numbers.Integral) and ax_it < 0:
                            sanitized_item[i] = aligned_shape[i] + ax_it
                        elif isinstance(ax_it, slice):
                            if (ax_it.start is not None and ax_it.start < 0) or (ax_it.stop is not None and ax_it.stop < 0):
                                new_start = aligned_shape[i] + ax_it.start if ax_it.start < 0 else ax_it.start
                                new_stop = aligned_shape[i] + ax_it.stop if ax_it.stop < 0 else ax_it.stop
                                sanitized_item[i] = slice(new_start, new_stop)
                    sanitized_item = tuple(sanitized_item)
                # Use sanitized item to slice meta.
                new_meta = self.meta.slice[sanitized_item]
            else:
                new_meta = copy.deepcopy(self.meta)

        return self.__class__(
            list(zip(new_keys, new_data)),
            aligned_axes=new_aligned_axes,
            meta=new_meta,
            sanitize_inputs=False
        )

    def _generate_collection_getitems(self, item):
        # There are 3 supported cases of the slice item: int, slice, tuple of ints and/or slices.
        # Compile appropriate slice items for each cube in the collection and
        # and drop any aligned axes that are sliced out.

        # First, define empty lists of slice items to be applied to each cube in collection.
        collection_items = [[slice(None)] * len(self[key].shape) for key in self]
        # Define empty list to hold aligned axes dropped by the slicing.
        drop_aligned_axes_indices = []

        # Case 1: int
        # First aligned axis is dropped.
        if isinstance(item, int):
            drop_aligned_axes_indices = [0]
            # Insert item to each cube's slice item.
            for i, key in enumerate(self):
                collection_items[i][self.aligned_axes[key][0]] = item

        # Case 2: slice
        elif isinstance(item, slice):
            # Insert item to each cube's slice item.
            for i, key in enumerate(self):
                collection_items[i][self.aligned_axes[key][0]] = item
            # Note that slice interval's of 1 result in an axis of length 1.
            # The axis is not dropped.

        # Case 3: tuple of ints/slices
        # Search sub-items within tuple for ints or 1-interval slices.
        elif isinstance(item, tuple):
            # Ensure item is not longer than number of aligned axes
            if len(item) > self.n_aligned_axes:
                raise IndexError("Too many indices")
            for i, axis_item in enumerate(item):
                if isinstance(axis_item, int):
                    drop_aligned_axes_indices.append(i)
                for j, key in enumerate(self):
                    collection_items[j][self.aligned_axes[key][i]] = axis_item

        else:
            raise TypeError(f"Unsupported slicing type: {axis_item}")

        # Use indices of dropped axes determine above to update aligned_axes
        # by removing any that have been dropped.
        drop_aligned_axes_indices = np.array(drop_aligned_axes_indices)
        new_aligned_axes = collection_utils._update_aligned_axes(
            drop_aligned_axes_indices, self.aligned_axes, self._first_key)

        return collection_items, new_aligned_axes

    def copy(self):
        # Aligned axes is not a required parameter and may be None
        aligned_axes = None if self.aligned_axes is None else tuple(self.aligned_axes.values())
        return self.__class__(self.items(), aligned_axes, meta=self.meta, sanitize_inputs=False)

    def setdefault(self):
        """Not supported by `~ndcube.NDCollection`"""
        raise NotImplementedError("NDCollection does not support setdefault.")

    def popitem(self):
        """Not supported by `~ndcube.NDCollection`"""
        raise NotImplementedError("NDCollection does not support popitem.")

    def pop(self, key):
        """
        Remove a member from the `~ndcube.NDCollection` and return it.

        Parameters
        ----------
        key: `str`
            The name of the member to remove and return.
        """
        # Extract desired cube from collection.
        popped_cube = super().pop(key)
        # Aligned axes is not a required parameter and may be None
        if self.aligned_axes is not None:
            # Delete corresponding aligned axes
            self.aligned_axes.pop(key)
        return popped_cube

    def update(self, *args):
        """
        Merges a new collection with current one replacing objects with common keys.

        Takes either a single input (`~ndcube.NDCollection`) or two inputs
        (sequence of key/value pairs and aligned axes associated with each key/value pair.
        """
        # If two inputs, inputs must be key_data_pairs and aligned_axes.
        if len(args) == 2:
            key_data_pairs = args[0]
            new_keys, new_data = zip(*key_data_pairs)
            new_aligned_axes = collection_utils._sanitize_aligned_axes(new_keys, new_data, args[1])
        else:  # If one arg given, input must be NDCollection.
            collection = args[0]
            new_keys = list(collection.keys())
            new_data = list(collection.values())
            key_data_pairs = zip(new_keys, new_data)
            new_aligned_axes = collection.aligned_axes

        # Check aligned axes of new inputs are compatible with those in self.
        # As they've already been sanitized, only one set of aligned axes need be checked.
        first_old_aligned_axes = self.aligned_axes[self._first_key] if self.aligned_axes is not None else None
        first_new_aligned_axes = new_aligned_axes[new_keys[0]] if new_aligned_axes is not None else None
        collection_utils.assert_aligned_axes_compatible(
            self[self._first_key].shape, new_data[0].shape,
            first_old_aligned_axes, first_new_aligned_axes
        )
        # Update collection
        super().update(key_data_pairs)
        if first_old_aligned_axes is not None:  # since the above assertion passed, if one aligned axes is not None, both are not None
            self.aligned_axes.update(new_aligned_axes)

    def __delitem__(self, key):
        super().__delitem__(key)
        self.aligned_axes.__delitem__(key)

    def __setitem__(self, key, value):
        raise NotImplementedError("NDCollection does not support __setitem__. "
                                  "Use NDCollection.update instead")




import pytest

from ndcube import utils
from ndcube.utils.sequence import SequenceItem

# sample data for tests
tuple_item0 = (0, slice(0, 3))
tuple_item1 = (slice(0, 2), slice(0, 3), slice(None))
tuple_item2 = (slice(3, 1, -1), slice(0, 3), slice(None))
tuple_item3 = (slice(4, None, -2), slice(0, 3), slice(None))

n_cubes = 4


@pytest.mark.parametrize(
    ("cube_like_index", "common_axis", "common_axis_lengths", "expected_seq_idx", "expected_common_idx"),
    [(3, 1, [4, 4], 0, 3),
     (3, 1, [2, 2], 1, 1)]
)
def test_cube_like_index_to_sequence_and_common_axis_indices(
        cube_like_index, common_axis, common_axis_lengths, expected_seq_idx, expected_common_idx):
    sequence_index, common_axis_index = \
        utils.sequence.cube_like_index_to_sequence_and_common_axis_indices(
            cube_like_index, common_axis, common_axis_lengths)
    assert sequence_index == expected_seq_idx
    assert common_axis_index == expected_common_idx


@pytest.mark.parametrize(
    ("item", "common_axis", "common_axis_lengths", "n_cube_dims", "expected_sequence_items"), [
        ((slice(None), slice(4, 6)), 1, [3, 3], 4,
         [SequenceItem(sequence_index=1, cube_item=slice(1, 3))]),

        ((slice(None), slice(None)), 1, [3, 3, 3], 4,
         [SequenceItem(sequence_index=0, cube_item=slice(0, None)),
          SequenceItem(sequence_index=1, cube_item=slice(None)),
          SequenceItem(sequence_index=2, cube_item=slice(None, 9))])]
)
def test_cube_like_tuple_item_to_sequence_items(
        item, common_axis, common_axis_lengths, n_cube_dims, expected_sequence_items):
    pass


def test_cube_like_tuple_item_to_sequence_items_error1():
    with pytest.raises(TypeError):
        utils.sequence.cube_like_tuple_item_to_sequence_items(1, 1, [2, 2], 3)


def test_cube_like_tuple_item_to_sequence_items_error2():
    with pytest.raises(ValueError):
        utils.sequence.cube_like_tuple_item_to_sequence_items((1, 1), 3, [2, 2], 4)


def test_cube_like_tuple_item_to_sequence_items_error3():
    with pytest.raises(TypeError):
        utils.sequence.cube_like_tuple_item_to_sequence_items((1, 1), 1, [2, 2], 3)



import functools

MISSING_MATPLOTLIB_ERROR_MSG = ("matplotlib cannot be imported, so the default plotting "
                                "functionality is disabled. Please install matplotlib")
MISSING_ANIMATORS_ERROR_MSG = ("mpl_animators cannot be imported, so the default plotting "
                               "functionality is disabled. Please install mpl_animators")

__all__ = ['MISSING_ANIMATORS_ERROR_MSG', 'MISSING_MATPLOTLIB_ERROR_MSG', 'PlotterDescriptor']


class PlotterDescriptor:
    def __init__(self, default_type=None):
        self._default_type = default_type

    def __set_name__(self, owner, name):
        """
        This function is called when the class the descriptor is attached to is initialized.

        The *class* and not the instance.
        """
        # property name is the name of the attribute on the parent class
        # pointing at an instance of this descriptor.
        self._property_name = name
        # attribute name is the name of the attribute on the parent class where
        # the data is stored.
        self._attribute_name = f"_{name}"
        plotter = self._resolve_default_type(raise_error=False)
        if plotter is not None and hasattr(plotter, "plot"):
            functools.update_wrapper(owner.plot, plotter.plot)

    def _resolve_default_type(self, raise_error=True):
        # We special case the default MatplotlibPlotter so that we can
        # delay the import of matplotlib until the plotter is first
        # accessed.
        if self._default_type in ("mpl_plotter", "mpl_sequence_plotter"):
            if self._default_type == "mpl_plotter":
                try:
                    from ndcube.visualization.mpl_plotter import MatplotlibPlotter
                    return MatplotlibPlotter
                except ImportError as e:
                    if raise_error:
                        raise ImportError(MISSING_MATPLOTLIB_ERROR_MSG) from e
            elif self._default_type == "mpl_sequence_plotter":
                try:
                    from ndcube.visualization.mpl_sequence_plotter import MatplotlibSequencePlotter
                    return MatplotlibSequencePlotter
                except ImportError as e:
                    if raise_error:
                        raise ImportError(MISSING_ANIMATORS_ERROR_MSG) from e
            return None

        if self._default_type is not None:
            return self._default_type

        # If we have no default type then just return None
        return None

    def __get__(self, obj, objtype=None):
        if obj is None:
            return None

        if getattr(obj, self._attribute_name, None) is None:
            plotter_type = self._resolve_default_type()
            if plotter_type is None:
                return None

            self.__set__(obj, plotter_type)

        return getattr(obj, self._attribute_name)

    def __set__(self, obj, value):
        if not isinstance(value, type):
            raise TypeError(
                "Plotter attribute can only be set with an uninitialised plotter object.")

        setattr(obj, self._attribute_name, value(obj))
        # here obj is the ndcube object and value is the plotter type
        # Get the instantiated plotter we just assigned to the ndcube
        plotter = getattr(obj, self._attribute_name)
        # If the plotter has a plot object then update the signature and
        # docstring of the cubes `plot()` method to match
        # Docstrings of methods aren't writeable so we copy to the underlying
        # function object instead
        if hasattr(plotter, "plot"):
            functools.update_wrapper(obj.plot.__func__, plotter.plot.__func__)




import numpy as np
import pytest

import astropy.wcs

from ndcube import NDCollection, NDCube, NDCubeSequence
from ndcube.tests import helpers

# Define some mock data
data0 = np.ones((3, 4, 5))
data1 = np.zeros((5, 3, 4))
data2 = data0 * 2

# Define WCS object for all cubes.
wcs_input_dict = {
    'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10, 'NAXIS1': 5,
    'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5, 'NAXIS2': 4,
    'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 3}
input_wcs = astropy.wcs.WCS(wcs_input_dict)

wcs_input_dict1 = {
    'CTYPE3': 'WAVE    ', 'CUNIT3': 'Angstrom', 'CDELT3': 0.2, 'CRPIX3': 0, 'CRVAL3': 10, 'NAXIS3': 5,
    'CTYPE1': 'HPLT-TAN', 'CUNIT1': 'deg', 'CDELT1': 0.5, 'CRPIX1': 2, 'CRVAL1': 0.5, 'NAXIS1': 4,
    'CTYPE2': 'HPLN-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.4, 'CRPIX2': 2, 'CRVAL2': 1, 'NAXIS2': 3}
input_wcs1 = astropy.wcs.WCS(wcs_input_dict1)

# Define cubes.
cube0 = NDCube(data0, input_wcs)
cube1 = NDCube(data1, input_wcs1)
cube2 = NDCube(data2, input_wcs)

# Define sequences.
sequence02 = NDCubeSequence([cube0, cube2])
sequence20 = NDCubeSequence([cube2, cube0])

# Define collections
aligned_axes = ((1, 2), (2, 0), (1, 2))
keys = ("cube0", "cube1", "cube2")
cube_coll_meta = helpers.ndmeta_et0_pr01((4, 5))
cube_collection = NDCollection([("cube0", cube0), ("cube1", cube1), ("cube2", cube2)], aligned_axes, meta=cube_coll_meta)
unaligned_collection = NDCollection([("cube0", cube0), ("cube1", cube1), ("cube2", cube2)], aligned_axes=None)
seq_collection = NDCollection([("seq0", sequence02), ("seq1", sequence20)], aligned_axes="all")


@pytest.mark.parametrize(('item', 'collection', 'expected'), [
    (0, cube_collection,
        NDCollection([("cube0", cube0[:, 0]), ("cube1", cube1[:, :, 0]), ("cube2", cube2[:, 0])],
            aligned_axes=((1,), (0,), (1,)), meta=cube_coll_meta.slice[0])),

    (slice(1, 3), cube_collection, NDCollection(
        [("cube0", cube0[:, 1:3]), ("cube1", cube1[:, :, 1:3]), ("cube2", cube2[:, 1:3])],
        aligned_axes=aligned_axes, meta=cube_coll_meta.slice[1:3])),

    (slice(-3, -1), cube_collection, NDCollection(
        [("cube0", cube0[:, -3:-1]), ("cube1", cube1[:, :, -3:-1]), ("cube2", cube2[:, -3:-1])],
        aligned_axes=aligned_axes, meta=cube_coll_meta.slice[1:3])),

    ((slice(None), slice(1, 2)), cube_collection, NDCollection(
        [("cube0", cube0[:, :, 1:2]), ("cube1", cube1[1:2]), ("cube2", cube2[:, :, 1:2])],
        aligned_axes=aligned_axes, meta=cube_coll_meta.slice[:, 1:2])),

    ((slice(2, 4), slice(-3, -1)), cube_collection, NDCollection(
        [("cube0", cube0[:, 2:4, -3:-1]), ("cube1", cube1[-3:-1, :, 2:4]),
            ("cube2", cube2[:, 2:4, -3:-1])], aligned_axes=aligned_axes, meta=cube_coll_meta.slice[2:4, 2:4])),

    ((0, 0), cube_collection, NDCollection(
        [("cube0", cube0[:, 0, 0]), ("cube1", cube1[0, :, 0]), ("cube2", cube2[:, 0, 0])],
        aligned_axes=None, meta=cube_coll_meta.slice[0, 0])),

    (("cube0", "cube2"), cube_collection, NDCollection(
        [("cube0", cube0), ("cube2", cube2)], aligned_axes=(aligned_axes[0], aligned_axes[2]), meta=cube_coll_meta)),

    (0, seq_collection, NDCollection([("seq0", sequence02[0]), ("seq1", sequence20[0])],
                                     aligned_axes=((0, 1, 2), (0, 1, 2)))),

    ((slice(None), 1, slice(1, 3)), seq_collection,
        NDCollection([("seq0", sequence02[:, 1, 1:3]), ("seq1", sequence20[:, 1, 1:3])],
                     aligned_axes=((0, 1, 2), (0, 1, 2))))
])
def test_collection_slicing(item, collection, expected):
    helpers.assert_collections_equal(collection[item], expected)


@pytest.mark.parametrize(('item', 'collection', 'expected'), [("cube1", cube_collection, cube1)])
def test_slice_cube_from_collection(item, collection, expected):
    helpers.assert_cubes_equal(collection[item], expected)


def test_collection_copy():
    helpers.assert_collections_equal(cube_collection.copy(), cube_collection)
    helpers.assert_collections_equal(unaligned_collection.copy(), unaligned_collection)


@pytest.mark.parametrize(('collection', 'popped_key', 'expected_popped', 'expected_collection'), [
    (cube_collection, "cube0", cube0, NDCollection([("cube1", cube1), ("cube2", cube2)],
                                                   aligned_axes=aligned_axes[1:])),
    (unaligned_collection, "cube0", cube0, NDCollection([("cube1", cube1), ("cube2", cube2)]))])
def test_collection_pop(collection, popped_key, expected_popped, expected_collection):
    popped_collection = collection.copy()
    output = popped_collection.pop(popped_key)
    helpers.assert_cubes_equal(output, expected_popped)
    helpers.assert_collections_equal(popped_collection, expected_collection)


@pytest.mark.parametrize(('collection', 'key', 'expected'), [
    (cube_collection, "cube0", NDCollection([("cube1", cube1), ("cube2", cube2)],
                                            aligned_axes=aligned_axes[1:]))])
def test_del_collection(collection, key, expected):
    del_collection = collection.copy()
    del del_collection[key]
    helpers.assert_collections_equal(del_collection, expected)


@pytest.mark.parametrize(('collection', 'key', 'data', 'aligned_axes', 'expected'), [
    (cube_collection, "cube1", cube2, aligned_axes[2], NDCollection(
        [("cube0", cube0), ("cube1", cube2), ("cube2", cube2)],
        aligned_axes=((1, 2), (1, 2), (1, 2)))),

    (cube_collection, "cube3", cube2, aligned_axes[2], NDCollection(
        [("cube0", cube0), ("cube1", cube1), ("cube2", cube2), ("cube3", cube2)],
        aligned_axes=((1, 2), (2, 0), (1, 2), (1, 2))))])
def test_collection_update_key_data_pair_input(collection, key, data, aligned_axes, expected):
    updated_collection = collection.copy()
    updated_collection.update([(key, data)], aligned_axes)
    helpers.assert_collections_equal(updated_collection, expected)


def test_collection_update_collecton_input():
    orig_collection = NDCollection([("cube0", cube0), ("cube1", cube1)], aligned_axes[:2])
    cube1_alt = NDCube(data1*2, input_wcs1)
    new_collection = NDCollection([("cube1", cube1_alt), ("cube2", cube2)], aligned_axes[1:])
    orig_collection.update(new_collection)
    expected = NDCollection([("cube0", cube0), ("cube1", cube1_alt), ("cube2", cube2)],
                            aligned_axes)
    helpers.assert_collections_equal(orig_collection, expected)


def test_collection_update_without_aligned_axes():
    orig_collection = NDCollection([("cube0", cube0), ("cube1", cube1)])
    new_collection = NDCollection([("cube2", cube2)])
    orig_collection.update(new_collection)

    expected = NDCollection([("cube0", cube0), ("cube1", cube1), ("cube2", cube2)])
    helpers.assert_collections_equal(orig_collection, expected)


@pytest.mark.parametrize(('collection', 'expected_aligned_dimensions'), [
    (cube_collection, [4, 5]),
    (seq_collection, [2, 3, 4, 5])])
def test_aligned_dimensions(collection, expected_aligned_dimensions):
    assert np.all(collection.aligned_dimensions == expected_aligned_dimensions)


@pytest.mark.parametrize(('collection', 'expected'), [
    (cube_collection, [('custom:pos.helioprojective.lat', 'custom:pos.helioprojective.lon'),
                       ('em.wl',)]),
    (seq_collection, [('meta.obs.sequence',),
                      ('custom:pos.helioprojective.lat', 'custom:pos.helioprojective.lon'),
                      ('custom:pos.helioprojective.lat', 'custom:pos.helioprojective.lon'),
                      ('em.wl',)])])
def test_aligned_axis_physical_types(collection, expected):
    output = collection.aligned_axis_physical_types
    assert len(output) == len(expected)
    for output_axis_types, expect_axis_types in zip(output, expected):
        assert set(output_axis_types) == set(expect_axis_types)



"""
==============================
Examples of cropping an NDCube
==============================

One of the powerful aspects of having coordinate-aware data stored as an
`~ndcube.NDCube` is the ability to crop and slice the data and coordinates in a
standardised and easy way.

For example, there may be a region of interest you would like to crop out along a certain dimension
of your cube. In this example, this method to slice an `~ndcube.NDCube` are illustrated.
"""
import numpy as np

import astropy.wcs
from astropy import units as u
from astropy.coordinates import SkyCoord, SpectralCoord

from sunpy.coordinates import frames

from ndcube import NDCube

##############################################################################
# Let's begin by creating an example `~ndcube.NDCube` object.
# For this case, we'll generate an `~ndcube.NDCube` that consists of 3 dimensions
# space-space-wavelength. This is analogous to an observation including images in multiple
# wavelengths.
# Lets first define a 3-D numpy array, and then define the WCS information that describes the
# data. We'll just create an array of random numbers, and a WCS which consists of the coordinate information
# which in this case will be in Helioprojective (i.e. an observation of the Sun) in latitude and longitude,
# and over several wavelengths in the range of 10.2 - 11 angstrom.

# Define the data of random numbers. Here the spatial dimensions are (45, 45) and the wavelength dimension is 5.
data = np.random.rand(5, 45, 45)
# Define the WCS
wcs = astropy.wcs.WCS(naxis=3)
wcs.wcs.ctype = 'HPLT-TAN', 'HPLN-TAN', "WAVE"
wcs.wcs.cunit = 'arcsec', 'arcsec', 'Angstrom'
wcs.wcs.cdelt = 10, 10, 0.2
wcs.wcs.crpix = 2, 2, 0
wcs.wcs.crval = 1, 1, 10
wcs.wcs.cname = 'HPC lat', 'HPC lon', 'wavelength'
# Instantiate the `~ndcube.NDCube`
example_cube = NDCube(data, wcs=wcs)

##############################################################################
# So we now have created an `~ndcube.NDCube` named ``example_cube``.
# You may have noticed that the order of the WCS is reversed to the array order - this
# is normal convention, and something to remember throughout.
# Now let's first inspect the cube.

##############################################################################
# Here we can inspect the cube by plotting it
example_cube.plot()

##############################################################################
# We can also inspect the shape of the cube:
example_cube.shape

##############################################################################
# We can also inspect the world coordinates for all array elements:
example_cube.axis_world_coords()

##############################################################################
# Slicing and cropping the cube
# -----------------------------
# An `~ndcube.NDCube` can be sliced and cropped both by array indexing (similar to the way a numpy
# array in indexed) or by real world coordinates. When we use array indices we say we
# are "slicing" the cube. When we use world coordinates we say we are "cropping" the cube.
# Let's begin by slicing by array index.

##############################################################################
# Slicing by array index
# ----------------------
# To slice the ``example_cube`` so that we extract only one wavelength, we can do:
# by indexing as such

sliced_cube = example_cube[1, :, :]
# here we can see we are left with a 2-D cube which is an image at one wavelength.
sliced_cube.shape

# We can also index a region of interest of the cube at a particular wavelength.
# Again note that we are slicing here based on the ``array`` index rather than cropping by
# real world value

sliced_cube = example_cube[1, 10:20, 20:40]
sliced_cube.shape

# Now we can inspect the sliced cube, and see it's now a smaller region of interest.
sliced_cube.plot()

##############################################################################
# Cropping cube using world coordinate values using :meth:`ndcube.NDCube.crop`
# ----------------------------------------------------------------------------
# In many cases it's more useful to crop a cube to a region of interest based
# on real world coordinates such as points in space or over some spectral
# range. This is achieved by the :meth:`ndcube.NDCube.crop` method which takes high-level astropy coordinate objects,
# such as `~astropy.coordinates.SkyCoord`. :meth:`ndcube.NDCube.crop` returns the smallest cube
# in array-index space that contains all the passed points.

##############################################################################
# Let's first define some points over which to crop the ``example_cube``. The points are
# defined as iterables of scalar high-level coordinate objects. We must provide the same
# number of objects in each tuple as required by the WCS to describe all the world axes.
# In this example, this means we need to provide a `~astropy.coordinates.SkyCoord` and
# `~astropy.coordinates.SpectralCoord` in each iterable. However, if we don't want to crop
# by one of the coordinate types, e.g. wavelength, we can replace the corresponding
# high-level coordinate object in each iterable with ``None``.
# Let's first define two points in space (lat and long) we want to crop but keep all wavelengths:

point1 = [SkyCoord(0*u.arcsec, 0*u.arcsec, frame=frames.Helioprojective), None]
point2 = [SkyCoord(200*u.arcsec, 100*u.arcsec, frame=frames.Helioprojective), None]

cropped_cube = example_cube.crop(point1, point2)

##############################################################################
# Similar to before, we can inspect the dimensions of the sliced cube via the shape property:

cropped_cube.shape

##############################################################################
# and we can visualize it:

cropped_cube.plot()

##############################################################################
# Now let's say we also want to crop out the image that includes a specific wavelength.
# Let's define a new point, and then include it with the first two we passed to :meth:`~ndcube.NDCube.crop`.

point3 = [None, SpectralCoord(10.2*u.angstrom)]

cropped_cube = example_cube.crop(point1, point2, point3)

##############################################################################
# we can inspect the dimensions of the cropped cube:

cropped_cube.shape

##############################################################################
# and again visualize it:

cropped_cube.plot()

# Now let's say we instead want to crop over a wavelength range.
# Let's define a new point, and then include it with the first two we passed to
# :meth:`~ndcube.NDCube.crop`.
point4 = [None, SpectralCoord(10.6*u.angstrom)]

cropped_cube = example_cube.crop(point1, point2, point3, point4)

##############################################################################
# Check dimensions:

cropped_cube.shape

##############################################################################
# Here we can just see how powerful this can be to easily crop over different world coordinates.
# In fact we can make this simpler still by combining the SkyCoords and SpectralCoords into
# just two points:

point5 = [SkyCoord(0*u.arcsec, 0*u.arcsec, frame=frames.Helioprojective), SpectralCoord(10.2*u.angstrom)]
point6 = [SkyCoord(200*u.arcsec, 100*u.arcsec, frame=frames.Helioprojective), SpectralCoord(10.6*u.angstrom)]

cropped_cube = example_cube.crop(point5, point6)
cropped_cube.shape



import pytest

from astropy.wcs.wcsapi.conftest import Celestial2DLowLevelWCS as ApyCelestial2DLowLevelWCS
from astropy.wcs.wcsapi.conftest import *  # NOQA


class Celestial2DLowLevelWCS(ApyCelestial2DLowLevelWCS):
    def __init__(self):
        self._pixel_bounds = (-1, 5), (1, 7)

    @property
    def pixel_bounds(self):
        return self._pixel_bounds

    @pixel_bounds.setter
    def pixel_bounds(self, val):
        self._pixel_bounds = val


@pytest.fixture
def celestial_2d_ape14_wcs():
    return Celestial2DLowLevelWCS()




"""
Helpers for testing ndcube.
"""
from pathlib import Path
from functools import wraps

import matplotlib as mpl
import matplotlib.pyplot as plt
import mpl_animators
import numpy as np
import pytest
from numpy.testing import assert_equal

import astropy
import astropy.units as u
from astropy.wcs.wcsapi import BaseHighLevelWCS
from astropy.wcs.wcsapi.fitswcs import SlicedFITSWCS
from astropy.wcs.wcsapi.low_level_api import BaseLowLevelWCS
from astropy.wcs.wcsapi.wrappers.sliced_wcs import sanitize_slices

from ndcube import NDCube, NDCubeSequence
from ndcube.meta import NDMeta

__all__ = [
    'assert_cubes_equal',
    'assert_cubesequences_equal',
    'assert_extra_coords_equal',
    'assert_metas_equal',
    'assert_wcs_are_equal',
    'figure_test',
    'get_hash_library_name',
]


def get_hash_library_name():
    """
    Generate the hash library name for this env.
    """
    ft2_version = f"{mpl.ft2font.__freetype_version__.replace('.', '')}"
    animators_version = "dev" if (("dev" in mpl_animators.__version__) or ("rc" in mpl_animators.__version__)) else mpl_animators.__version__.replace('.', '')
    mpl_version = "dev" if (("dev" in mpl.__version__) or ("rc" in mpl.__version__)) else mpl.__version__.replace('.', '')
    astropy_version = "dev" if (("dev" in astropy.__version__) or ("rc" in astropy.__version__)) else astropy.__version__.replace('.', '')
    return f"figure_hashes_mpl_{mpl_version}_ft_{ft2_version}_astropy_{astropy_version}_animators_{animators_version}.json"


def figure_test(test_function):
    """
    A decorator for a test that verifies the hash of the current figure or the
    returned figure, with the name of the test function as the hash identifier
    in the library. A PNG is also created in the 'result_image' directory,
    which is created on the current path.

    All such decorated tests are marked with ``pytest.mark.mpl_image`` for convenient filtering.
    """
    hash_library_name = get_hash_library_name()
    hash_library_file = Path(__file__).parent / ".." / "visualization" / "tests" / hash_library_name

    @pytest.mark.remote_data
    @pytest.mark.mpl_image_compare(hash_library=hash_library_file.resolve(),
                                   savefig_kwargs={'metadata': {'Software': None}},
                                   style='default')
    @wraps(test_function)
    def test_wrapper(*args, **kwargs):
        ret = test_function(*args, **kwargs)
        if ret is None:
            ret = plt.gcf()
        return ret
    return test_wrapper


def assert_extra_coords_equal(test_input, extra_coords):
    assert set(test_input.keys()) == set(extra_coords.keys())
    if extra_coords._lookup_tables is None:
        assert test_input._lookup_tables is None
    for ec_idx, key in enumerate(extra_coords.keys()):
        test_idx = np.where(np.asarray(test_input.keys()) == key)[0][0]
        assert test_input.mapping[test_idx] == extra_coords.mapping[ec_idx]
        if extra_coords._lookup_tables is not None:
            test_table = test_input._lookup_tables[test_idx][1].table
            ec_table = extra_coords._lookup_tables[ec_idx][1].table
            if not isinstance(ec_table, tuple):
                test_table = (test_table,)
                ec_table = (ec_table,)
            for test_tab, ec_tab in zip(test_table, ec_table):
                if ec_tab.isscalar:
                    assert test_tab == ec_tab
                else:
                    assert all(test_tab == ec_tab)
    if extra_coords._wcs is None:
        assert test_input._wcs is None
    else:
        assert_wcs_are_equal(test_input._wcs, extra_coords._wcs)


def assert_metas_equal(test_input, expected_output):
    if type(test_input) is not type(expected_output):
        raise AssertionError(
            "input and expected are of different type. "
            f"input: {type(test_input)}; expected: {type(expected_output)}")
    multi_element_msg = "more than one element is ambiguous"
    if isinstance(test_input, NDMeta) and isinstance(expected_output, NDMeta):
        assert test_input.keys() == expected_output.keys()

        if test_input.data_shape is None or expected_output.data_shape is None:
            assert test_input.data_shape == expected_output.data_shape
        else:
            assert np.allclose(test_input.data_shape, expected_output.data_shape)

        for test_value, expected_value in zip(test_input.values(), expected_output.values()):
            try:
                assert test_value == expected_value
            except ValueError as err:  # noqa: PERF203
                if multi_element_msg in err.args[0]:
                    assert np.allclose(test_value, expected_value)
        for key in test_input.axes.keys():
            assert all(test_input.axes[key] == expected_output.axes[key])
    else:
        if not (test_input is None and expected_output is None):
            assert test_input.keys() == expected_output.keys()
            for key in list(test_input.keys()):
                assert test_input[key] == expected_output[key]


def assert_cubes_equal(test_input, expected_cube, check_data=True):
    assert isinstance(test_input, type(expected_cube))
    assert np.all(test_input.mask == expected_cube.mask)
    if check_data:
        np.testing.assert_array_equal(test_input.data, expected_cube.data)
    assert_wcs_are_equal(test_input.wcs, expected_cube.wcs)
    if test_input.uncertainty:
        assert test_input.uncertainty.array.shape == expected_cube.uncertainty.array.shape
    assert np.all(test_input.shape == expected_cube.shape)
    assert_metas_equal(test_input.meta, expected_cube.meta)
    if type(test_input.extra_coords) is not type(expected_cube.extra_coords):
        raise AssertionError(f"NDCube extra_coords not of same type: "
                             f"{type(test_input.extra_coords)} != {type(expected_cube.extra_coords)}")
    if test_input.extra_coords is not None:
        assert_extra_coords_equal(test_input.extra_coords, expected_cube.extra_coords)


def assert_cubesequences_equal(test_input, expected_sequence, check_data=True):
    assert isinstance(test_input, type(expected_sequence))
    assert_metas_equal(test_input.meta, expected_sequence.meta)
    assert test_input._common_axis == expected_sequence._common_axis
    for i, cube in enumerate(test_input.data):
        assert_cubes_equal(cube, expected_sequence.data[i], check_data=check_data)


def assert_wcs_are_equal(wcs1, wcs2):
    """
    Assert function for testing two wcs object.

    Used in testing NDCube.
    Also checks if both the wcs objects are instance
    of `~astropy.wcs.wcsapi.SlicedLowLevelWCS`.
    """

    if not isinstance(wcs1, BaseLowLevelWCS):
        wcs1 = wcs1.low_level_wcs
    if not isinstance(wcs2, BaseLowLevelWCS):
        wcs2 = wcs2.low_level_wcs
    # Check the APE14 attributes of both the WCS
    assert wcs1.pixel_n_dim == wcs2.pixel_n_dim
    assert wcs1.world_n_dim == wcs2.world_n_dim
    assert wcs1.array_shape == wcs2.array_shape
    assert wcs1.pixel_shape == wcs2.pixel_shape
    assert wcs1.world_axis_physical_types == wcs2.world_axis_physical_types
    assert wcs1.world_axis_units == wcs2.world_axis_units
    assert_equal(wcs1.axis_correlation_matrix, wcs2.axis_correlation_matrix)
    assert wcs1.pixel_bounds == wcs2.pixel_bounds
    if wcs1.pixel_shape is not None:
        random_idx = np.random.randint(wcs1.pixel_shape,size=[10,wcs1.pixel_n_dim])
        # SlicedLowLevelWCS vs BaseHighLevelWCS don't have the same pixel_to_world method
        low_level_wcs1 = wcs1.low_level_wcs if isinstance(wcs1, BaseHighLevelWCS) else wcs1
        low_level_wcs2 = wcs2.low_level_wcs if isinstance(wcs2, BaseHighLevelWCS) else wcs2
        np.testing.assert_array_equal(low_level_wcs1.pixel_to_world_values(*random_idx.T), low_level_wcs2.pixel_to_world_values(*random_idx.T))

def create_sliced_wcs(wcs, item, dim):
    """
    Creates a sliced `SlicedFITSWCS` object from the given slice item
    """

    # Sanitize the slices
    item = sanitize_slices(item, dim)
    return SlicedFITSWCS(wcs, item)


def assert_collections_equal(collection1, collection2, check_data=True):
    assert collection1.keys() == collection2.keys()
    assert collection1.aligned_axes == collection2.aligned_axes
    for cube1, cube2 in zip(collection1.values(), collection2.values()):
        # Check cubes are same type.
        assert type(cube1) is type(cube2)
        if isinstance(cube1, NDCube):
            assert_cubes_equal(cube1, cube2, check_data=check_data)
        elif isinstance(cube1, NDCubeSequence):
            assert_cubesequences_equal(cube1, cube2, check_data=check_data)
        else:
            raise TypeError(f"Unsupported Type in NDCollection: {type(cube1)}")

def ndmeta_et0_pr01(shape):
    return NDMeta({"salutation": "hello",
                   "exposure time": u.Quantity([2.] * shape[0], unit=u.s),
                   "pixel response": (100 * np.ones((shape[0], shape[1]), dtype=float)) * u.percent},
                   axes={"exposure time": 0, "pixel response": (0, 1)}, data_shape=shape)


def ndmeta_et0_pr02(shape):
    return NDMeta({"salutation": "hello",
                   "exposure time": u.Quantity([2.] * shape[0], unit=u.s),
                   "pixel response": (100 * np.ones((shape[0], shape[2]), dtype=float)) * u.percent},
                   axes={"exposure time": 0, "pixel response": (0, 2)}, data_shape=shape)



import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal

from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.wcsapi import SlicedLowLevelWCS

from ndcube.wcs.tools import unwrap_wcs_to_fitswcs
from ndcube.wcs.wrappers import ResampledLowLevelWCS


def test_unwrap_wcs_to_fitswcs():
    # Build FITS-WCS and wrap it in different operations.
    time_ref = Time("2000-01-01T00:00:00", scale="utc", format="fits")
    header = {
        "CTYPE1": "TIME", "CTYPE2": "WAVE", "CTYPE3": "HPLT-TAN", "CTYPE4": "HPLN-TAN",
        "CUNIT1": "s", "CUNIT2": "Angstrom", "CUNIT3": "deg", "CUNIT4": "deg",
        "CDELT1": 600, "CDELT2": 0.2, "CDELT3": 0.5, "CDELT4": 0.4,
        "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 2, "CRPIX4": 2,
        "CRVAL1": 0, "CRVAL2": 10, "CRVAL3": 0.5, "CRVAL4": 1,
        "CNAME1": "time", "CNAME2": "wavelength", "CNAME3": "HPC lat", "CNAME4": "HPC lon",
        "NAXIS1": 5, "NAXIS2": 9, "NAXIS3": 4, "NAXIS4": 4,
        "DATEREF": time_ref.fits}
    orig_wcs = WCS(header)
    # Slice WCS
    wcs1 = SlicedLowLevelWCS(orig_wcs, (0, 0, slice(None), slice(1, None)))  # numpy order
    # Resample WCS
    wcs2 = ResampledLowLevelWCS(wcs1, [2, 3], offset=[0.5, 1])  # WCS order
    # Slice WCS again
    wcs3 = SlicedLowLevelWCS(wcs2, (slice(0, 2), slice(1, 2)))  # numpy order
    # Reconstruct fitswcs
    output_wcs, dropped_data_dimensions = unwrap_wcs_to_fitswcs(wcs3)
    # Assert output_wcs is correct
    assert_array_equal(dropped_data_dimensions, np.array([True, True, False, False]))
    assert isinstance(output_wcs, WCS)
    assert output_wcs._naxis == [1, 2, 1, 1]
    assert list(output_wcs.wcs.ctype) == ['TIME', 'WAVE', 'HPLT-TAN', 'HPLN-TAN']
    world_values = output_wcs.array_index_to_world_values([0], [0], [0, 1], [0])
    assert_array_almost_equal(world_values[0][0], np.array([2700]))
    assert_array_almost_equal(world_values[1], np.array([1.04e-09, 1.10e-09]))
    assert_array_almost_equal(world_values[2][0], np.array([1.26915033e-05]))
    assert_array_almost_equal(world_values[3][0], np.array([0.60002173]))



"""
A sphinx directive to execute a code block with context, and to provide the
user a way to see that context.
"""

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.directives.code import CodeBlock


class ExpandingCodeBlock(CodeBlock):
    """
    A code-block directive which is wrapped in a HTML details tag.

    It behaves like the code-block directive, with the addition of a
    ``:summary:`` option, which sets the unexpanded text.
    """
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = {
        'force': directives.flag,
        'linenos': directives.flag,
        'dedent': int,
        'lineno-start': int,
        'emphasize-lines': directives.unchanged_required,
        'caption': directives.unchanged_required,
        'class': directives.class_option,
        'name': directives.unchanged,
        'summary': directives.unchanged_required,
    }

    def run(self):
        source, lineno = self.state_machine.get_source_and_line(self.lineno)

        summary_text = self.options.get('summary', 'Show setup code')

        opening_details = f"""\
        <details>
          <summary style="display: list-item; cursor: pointer;">{summary_text}</summary>
        """
        open_raw_node = nodes.raw('', opening_details, format='html')
        open_raw_node.source, open_raw_node.line = source, lineno

        close_raw_node = nodes.raw('', "</details>", format='html')
        close_raw_node.source, close_raw_node.line = source, lineno

        literal = super().run()[0]

        return [open_raw_node, literal, close_raw_node]


def setup(app):
    app.add_directive('expanding-code-block', ExpandingCodeBlock)

    return {'parallel_read_safe': True, 'parallel_write_safe': True}



"""
================================================
Combining a celestial WCS with a wavelength axis
================================================

The goal of this example is to construct a spectral-image cube of AIA images at different wavelengths.

This will showcase how to add an arbitrarily spaced wavelength dimension
to a celestial WCS.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map

from ndcube import NDCube
from ndcube.extra_coords import QuantityTableCoordinate
from ndcube.wcs.wrappers import CompoundLowLevelWCS

#############################################################################
# We will use the sample data that ``sunpy`` provides to construct a sequence of AIA
# image files for different wavelengths using `sunpy.map.Map`.

aia_files = [sunpy.data.sample.AIA_094_IMAGE,
             sunpy.data.sample.AIA_131_IMAGE,
             sunpy.data.sample.AIA_171_IMAGE,
             sunpy.data.sample.AIA_193_IMAGE,
             sunpy.data.sample.AIA_211_IMAGE,
             sunpy.data.sample.AIA_304_IMAGE,
             sunpy.data.sample.AIA_335_IMAGE,
             sunpy.data.sample.AIA_1600_IMAGE]
# `sequence=True` causes a sequence of maps to be returned, one for each image file.
sequence_of_maps = sunpy.map.Map(aia_files, sequence=True)
# Sort the maps in the sequence in order of wavelength.
sequence_of_maps.maps = sorted(sequence_of_maps.maps, key=lambda m: m.wavelength)

#############################################################################
# Using an `astropy.units.Quantity` of the wavelengths of the images, we can construct
# a 1D lookup-table WCS via `.QuantityTableCoordinate`.
# This is then combined with the celestial WCS into a single 3D WCS via `.CompoundLowLevelWCS`.

wavelengths = u.Quantity([m.wavelength for m in sequence_of_maps])
wavelengths_wcs = QuantityTableCoordinate(wavelengths, physical_types="em.wl", names="wavelength").wcs
cube_wcs = CompoundLowLevelWCS(wavelengths_wcs, sequence_of_maps[0].wcs)

#############################################################################
# Combine the new 3D WCS with the stack of AIA images using `ndcube.NDCube`.
# Note that because we set the wavelength to the first axis
# in the WCS (cube_wcs), the final data cube is stacked such
# that wavelength corresponds to the array axis which is last.
# This is due to the convention that WCS axis ordering is reversed
# compared to data array axis ordering.

my_cube = NDCube(sequence_of_maps.as_array(), wcs=cube_wcs)
# Produce an interactive plot of the spectral-image stack.
my_cube.plot(plot_axes=['y', 'x', None])

plt.show()
